###################################################
## AUXILIARY FUNCTIONS
###################################################

library("gtools")

# Relations between nodes (i,j) or groups (q,l) are stored in vectors, 
# whoses indexes dependent on whether the graph is directed or undirected.
# 
# Directed case : 
#   Number of dyads N= n*(n-1)
#   The (i,j) relation (i\neq j) is stored in variable with index (i-1)*(n-1)+j-(i<j)
#   Type is an integer between 1 and N corresponding to a dyad (i,j)
#   The (q,l) group is stored in variable with index (q-1)*Q+l 
# Undirected case : 
#   Number of dyads N= n*(n-1)/2
#   The (i,j) relation (i < j) is stored in variable with index (2*n-i)*(i-1)/2 +j-i
#   Type is an integer between 1 and N corresponding to a dyad (i,j)
#   The (q,l) group with q<=l is stored in variable with index (2*Q-q+2)*(q-1)/2 +l-q+1
# 


###################################################
##   HANDLING OF PROCESS INDICES (i,j)
###################################################

####
# Numbering events (i,j) 
####

which.dyads <- function(i,j,n,directed){ 
  if (sum((i>n) | (j>n))>0){
    stop("Your index is out of range")
  }
  if (directed){#directed
    dyads = (i-1)*(n-1)+j-(i<j)
  } else {dyads <- c(0,cumsum((n-1):1))[pmin(i,j)] + abs(j-i)}
  return(dyads)  
}


which.index.all <- function(n,directed=TRUE){ 
  N <- if (directed) n*(n-1) else n*(n-1)/2
  index <- matrix(0,N,2)
  if (directed){ # directed
 	  index[,1] <- rep(1:n,each=n-1)
	  k <- (1:n^2)[-seq(1,n^2,by=n+1)]
	  index[,2] <- rep(1:n,n)[k]
  }else { # undirected
    index[,1] <- rep(1:(n-1),times=(n-1):1)
    toto <- c()
    for (k in 1:(n-2)){
      toto <- c(toto,k*(n-1)+ 1:k)
    }
    index[,2] <- rep(2:n,n-1)[-toto]
  }
	return(index)
}
 


###################################################
##   HANDLING OF GROUP INDICES (q,l)
###################################################

####
# Gives the index in 1..Q^2 (directed) or 1..Q*(Q+1)/2 (undirected) that corresponds to group (q,l)
# works also for vectors of indices q and l
####

indice <- function(q,l,Q,directed=TRUE){
  if (sum((q>Q) | (l>Q))>0){
    stop("Your index is out of range")
  }
  if (directed){
    index <- (q-1)*Q+l
  } else { # undirected
    qp <- pmin(q,l)
    lp <- pmax(q,l)
    index <- (2*Q-qp+2)*(qp-1)/2 +lp-qp+1
  }
	return(index)	
}


## This function is the inverse of the numberings 
# {(q,l), q,l} into {1,...,Q^2} for the directed case
# {(q,l), q<= l} into {1,...,Q*(Q+1)/2} for the undirected case
find.q_l <- function(ind_ql, Q,directed=TRUE){
  # test coherence
  nb.groups.ql <- if (directed)  Q^2 else Q*(Q+1)/2
  if (ind_ql > nb.groups.ql) stop("Your index is out of range")
  
  if (directed){ # directed
    q <- ceiling(ind_ql/Q)
    l <- ind_ql - Q*(q-1) 
  }else{ # undirected
    w <- cumsum(Q:1)
    q <- which.max(ind_ql<=w)
    w <- c(0,w)
    l <- ind_ql - w[q] + q - 1
  }
  return(c(q,l))
}



## This function is the inverse of the numbering {(q,l), q<l} into {1,...,Q*(Q-1)/2}
# used only in tau.down_Q
find.ql.diff <- function(ind_ql,Q){
  if (ind_ql > Q*(Q-1)/2){
    stop("Your index is out of range")
  }
  w <- cumsum((Q-1):1)
  q <- which.max(ind_ql<=w)
  w <- c(0,w)
  l <- ind_ql - w[q] + q 
  return(c(q,l))
}

###################################################
##   HANDLING OF VALUES OF TAU
###################################################

####
# avoid values of tau to be exactly 0 and exactly 1
####

correct.tau <- function(tau){
	  tau <- pmin(tau,.Machine$double.xmax)
    tau <- pmax(tau,.Machine$double.xmin)
    tau <- tau/sum(tau)
    tau <- pmin(tau,1-1e-7)
    tau <- pmax(tau,1e-7) 
    tau <- tau/sum(tau)	
	
	return(tau)
}

  
###################################################
##   COUNTS associated with data
###################################################

####
# compute statistic N(i,j)k = number of events of process (i,j) during k-th subinterval 
# from data (= list with $type.seq, $time.seq)
# K is the size of the regular partition (= number of intervals)
####

statistics <- function(data,n,K,directed=TRUE){ 
  partition <- data$Time*seq(1/K,1,by=1/K)
  N <- if (directed)  n*(n-1) else n*(n-1)/2
  Nijk <- matrix(0,N,K)
  for (ind in 1:N){
    events.ij <- data$time.seq[data$type.seq==ind]
    nb.ij <- length(events.ij)
    if (nb.ij>0){	
      counts <- apply(matrix(events.ij,nrow=K,ncol=nb.ij,byrow=TRUE)<matrix(partition,nrow=K,ncol=nb.ij),1,sum)
      Nijk[ind,] <- if (K>1) counts-c(0,counts[1:(K-1)]) else counts
      }
    }
  return(Nijk)	
}

 