####################################
#### INITIALIZATION OF VEM
####################################

library("gtools")

############
# list of initial values for tau
############
# same function whatever directed or undirected case
# l_part - maximal level d for dyadic partitions of time interval [0,T] used for kmeans initializations
#          Algorithm takes partition up to depth 2^d
#          Explore partitions [0,T], [0,T/2], [T/2,T], ... [0,T/2^d], ...[(2^d-1)T/2^d,T]
#          Total number of partitions npart= 2^(l_part +1) -1
# n_perturb - number of different perturbations on k-means result
# perc_perturb - percentage of labels that are to be perturbed (= randomly switched)
# n_random - nb of completely random initial points
# The (maximal) total number of initializations is npart*(1+n_perturb) +n_random

initial.tau <- function(data,n,Q,l_part,n_perturb,perc_perturb,n_random,directed){
  nbstart <- (2^(l_part+1)-1)*(1+n_perturb)
  init.values <- list()  
  if (Q==1)
    init.values <- list(matrix(1,1,n))
  else{
    if (nbstart>0){
      nbswitch <- min(c(max(c(round(n*perc_perturb),2)),n))
      Dmax <- ncol(data$Nijk)
#      nn <- min(c(2^l_part,Dmax))
#      n_levels <- floor(log2(nn))
      n_levels <- min(c(l_part,log2(Dmax)))
      for (d in 0:n_levels){
        counts <- data$Nijk
        while(ncol(counts)>2^d){
          counts <- as.matrix(counts[,seq(1,ncol(counts)-1,by=2)]+counts[,seq(2,ncol(counts),by=2)])
        }
        for (k in 1:2^d){
            if (sum(counts[,k]>0)>=Q){ # condition for kmeans to cluster into Q groups
              tau <- sbm_kmeans(counts[,k],n,Q,directed)
              tau <- apply(tau,2,correct.tau)
              init.values <- append(init.values,list(tau))
              # perturbation by label switching
              if (n_perturb>0){
                for (l in 1:n_perturb){
                  tau.copy <- tau
                  label2switch <- sample(1:n,nbswitch)
                  tau.copy[,label2switch] <- tau[,sample(label2switch,nbswitch)]
                  tau.copy <- apply(tau.copy,2,correct.tau)
                  init.values <- append(init.values,list(tau.copy))
                }
              }
            } 
        }# end loop for k in 1:2^d
      } # end loop for d
    }
    if (n_random>0){
      for (k in (nbstart+1):(nbstart+n_random)){
        tau <- matrix(rdirichlet(n,rep(1,Q)),Q,n,byrow=T)
        tau <- apply(tau,2,correct.tau)
        init.values <- append(init.values,list(tau))
      }
    }
  } # end else Q!=1
  return(init.values)
}


# sparse setup - rho parameter
initial.taurho <- function(tau,data,n,Q,directed=TRUE){
  if (directed){ 
    N <- n*(n-1)
    N_Q <- Q^2
  }
  else{ # undirected
    N <- n*(n-1)/2
    N_Q <- Q*(Q+1)/2
  }
  ind.all <- which.index.all(n,directed) 
  Nij <- apply(data$Nijk,1,sum)
  rho <- 1:N_Q
  for (ind.ql in 1:N_Q){
    if (Q==1)	
      tauql <- rep(1,N)
    else{
      ql <- find.q_l(ind.ql,Q,directed)
      q <- ql[1]
      l <- ql[2]
      if (directed) # directed case
        tauql <- tau[q, ind.all[,1]]*tau[l, ind.all[,2]] # N-vector
      else{ # undirected
        if ((q==l))
          tauql <- tau[q, ind.all[,1]]*tau[l, ind.all[,2]] 
        else
          tauql <- tau[q, ind.all[,1]]*tau[l, ind.all[,2]] + tau[q, ind.all[,2]]*tau[l, ind.all[,1]] 
        }
      }
      nb.procql.connect <- sum((Nij>0)*tauql)
      Yql <- sum(tauql)
      beta <- nb.procql.connect/Yql
      AqlT <- sum(Nij*tauql)/nb.procql.connect  
      rho[ind.ql] <- 1/((1/beta-1)*exp(AqlT)+1)
    }
  return(list(tau=tau,rho=rho))	
}

#######
## Construct initial tau with Q groups from value obtained at Q-1 groups 
#######
tau.up_Q <- function(tau,n_perturb=1){
  Q <- dim(tau)[1]
  n <- dim(tau)[2]
  tau.list <- list() 
  
  if (Q==1){
    tau.list <- list(matrix(1/2,2,n))
  } else{ # Q>=2
    if (n_perturb>=Q){ # split all the groups once
      for (q in 1:Q){
        tau.new <- tau
        tau.new[q,] <- tau.new[q,]/2
        tau.new <- rbind(tau.new,tau.new[q,])
        tau.list <- append(tau.list,list(tau.new))
      }
    } else { # always split the largest group 
      largest.gp <- which.max(apply(tau,1,sum))
      tau.new <- tau
      tau.new[largest.gp,] <- tau.new[largest.gp,]/2
      tau.new <- rbind(tau.new,tau.new[largest.gp,])
      tau.list <- append(tau.list,list(tau.new))
      if (n_perturb >=2){
        # and split the component with largest entropy
        ent <- tau*log(tau)
        ent[is.na(ent)] <- 0
        ent.gp <- which.min(apply(ent,1,sum))
        tau.new <- tau
        tau.new[ent.gp,] <- tau.new[ent.gp,]/2
        tau.new <- rbind(tau.new,tau.new[ent.gp,])
        tau.list <- append(tau.list,list(tau.new))
      }
      # And split a number n_perturb-2 of groups which are randomly chosen
      if (n_perturb >2){
        split.gp <- sample((1:Q)[-c(largest.gp,ent.gp)],n_perturb-2,replace=F)
        for (q in split.gp){
          tau.new <- tau
          tau.new[q,] <- tau.new[q,]/2
          tau.new <- rbind(tau.new,tau.new[q,])
          tau.list <- append(tau.list,list(tau.new))
         }  
       }
     } # end if/else n_perturb>=Q
   } # end if/else Q==1
  
  return(tau.list)
}

#######
## Construct initial tau with Q groups from value obtained at Q+1 groups 
#######
tau.down_Q <- function(tau,n_perturb=1){
  Q <- dim(tau)[1]
  n <- dim(tau)[2]
  tau.list <- list() 

  if (Q==2){
    tau.list <- list(matrix(1,1,n))
  } else {
    if (n_perturb>=Q*(Q-1)/2){ # merge all the possible pairs of groups (q,l) with q<l
      for (q in 1:(Q-1)){
        for (l in ((q+1):Q)){
          tau.new <- tau
          tau.new <- rbind(tau.new[-c(q,l),],tau.new[q,]+tau.new[l,])
          tau.list <- append(tau.list,list(tau.new))
        }
      }
    } else {
      # start by merging the 2 components with largest entropies
      ent <- tau*log(tau)
      ent[is.na(ent)] <- 0
      entropy.per.gp <- apply(ent,1,sum)
      q <- order(entropy.per.gp)[1]
      l <- order(entropy.per.gp)[2]
      tau.new <- tau
      tau.new <- rbind(tau.new[-c(q,l),],tau.new[q,]+tau.new[l,])          
      tau.list <- append(tau.list,list(tau.new))
      # merge a number n_perturb-1 of pairs of groups (q,l) q<l which are randomly chosen
      merge.gp <- sample(1:(Q*(Q-1)/2),n_perturb-1,replace=F)
      for (ind.ql in merge.gp){
        tau.new <- tau
        q <- find.ql.diff(ind.ql,Q)[1]  # this is defined in auxiliary.R
        l <- find.ql.diff(ind.ql,Q)[2]  
        tau.new <- rbind(tau.new[-c(q,l),],tau.new[q,]+tau.new[l,])          
        tau.list <- append(tau.list,list(tau.new))
      }
    } # end if/else n_perturb>=Q*(Q-1)/2
  } # end if/else Q==2
  return(tau.list)
}



######################
####   K-MEANS FOR SBM
######################

####
## function for k-means		
####

class.ind<-function (cl){
  ## INPUT
  ##    cl    : label list of nodes
  ## OUPUT
  ##    x    : class indicator matrix
  ## class.ind builds a class indicator matrix
  ## from a list of labels
  ## nq  <- rmultinom(1, size=7, prob = c(1/2, 1/8, 1/8))
  ## cl  <- rep(1:3, nq)
  ## cl = [1] 1 1 1 1 2 2 3
  ## class.ind(cl)
  ##      1 2 3
  ## [1,] 1 0 0
  ## [2,] 1 0 0
  ## [3,] 1 0 0
  ## [4,] 1 0 0
  ## [5,] 0 1 0
  ## [6,] 0 1 0
  ## [7,] 0 0 1
  
  n         <- length(cl)
  cl         <- as.factor(cl)
  x         <- matrix(0, n, length(levels(cl)))
  x[(1:n) + n * (unclass(cl) - 1)] <- 1
  dimnames(x) <- list(names(cl), levels(cl))
  x
}


####
## k-means for SBM
####

sbm_kmeans <- function(counts,n,Q,directed){
  X   <- matrix(0,n,n)
  if (!directed){
    k <- 1
    for (i in 1:(n-1)){
      X[i,(i+1):n] <- counts[k:(k+n-i-1)]
      k <- k+n-i
    }
    X <- X + t(X)
  }
  else{
    for (i in 1:n){
      X[i,-i] <- counts[((i-1)*(n-1)+1):(i*(n-1))]
    }
  }
  
  try.km <- try(kmeans(X,Q,nstart=50),silent=T)
  if (is.list(try.km)) {
    tau <- t(class.ind(try.km$cluster))
  } else {tau <- matrix(rdirichlet(n,rep(1,Q)),Q,n,byrow=T)} 
  tau <- apply(tau,2,correct.tau)
  
  return(tau)
}

