###################################################
## MODEL SELECTION BY ICL  &
## ANALYSIS OF RESULTS 
###################################################

source("vem_dynppsbm.R")
# install.packages("clue")
library("clue")


###################################################
## Selects the number of groups with ICL
###################################################
# data - list with 2 components: 
#   $Time - [0,data$Time] is the total time interval of observation 
#   $Nijk - data matrix with counts per process N_ij and sub-intervals ; 
# sol.hist.sauv : list of size Qmax-Qmin+1 obtained from running dynppSBM(data,n,Qmin,Qmax,method='hist') 

model.selection_Q <- function(data,n,Qmin=1,Qmax,directed=TRUE,sparse=FALSE,sol.hist.sauv){
  N <- if (directed) n*(n-1) else n*(n-1)/2
  
  ICLbest <- -Inf
  sol.Qbest <- NA
  all.ICL <- rep(NA,Qmax-Qmin+1)
  all.compl.log.likelihood <- rep(NA,Qmax-Qmin+1)
  all.J <- rep(NA,Qmax-Qmin+1)
  all.pen  <- rep(NA,Qmax-Qmin+1)
  for (Qcand in Qmin:Qmax){
    cat("----Qcand----",Qcand,"\n")
    sol.hist <-sol.hist.sauv[[Qcand]]
    all.J[Qcand-Qmin+1]<-sol.hist$J
    
    # arguments for complete.log.lik 
    VE <- list(tau=sol.hist$tau,rho=sol.hist$rho)
    no.obs <- (apply(data$Nijk,1,sum)==0)   
    Nij <- rowSums(data$Nijk)
    # compute sum_rho_obs
    N_Q <- if (directed) Qcand^2 else Qcand*(Qcand+1)/2
    sum_rhotau_obs <- rep(0,N_Q) 
    if (sparse){
      ind.all <- which.index.all(n,directed) 
      ind.ql <- 0
      for (q in 1:Qcand){
        ind.l <- if (directed) 1:Qcand else q:Qcand
        for (l in ind.l){
          ind.ql <- ind.ql + 1
          if (Qcand==1)	{
            tauql <- rep(1,N)
          }
          else {
            if (directed){ # directed case 
              tauql <- sol.hist$tau[q, ind.all[,1]]*sol.hist$tau[l, ind.all[,2]] # N-vector
            } else { # undirected case
              if ((q==l))
                tauql <- sol.hist$tau[q, ind.all[,1]]*sol.hist$tau[l, ind.all[,2]] # N-vector
              else
                tauql <- sol.hist$tau[q, ind.all[,1]]*sol.hist$tau[l, ind.all[,2]] + sol.hist$tau[q, ind.all[,2]]*sol.hist$tau[l, ind.all[,1]] 
            }
          } # end else Q!=1
          rhoql.vec <- rep(1,N)
          rhoql.vec[no.obs] <- sol.hist$rho[ind.ql]
          sum_rhotau_obs[ind.ql] <- sum(tauql * rhoql.vec * Nij) 
        } # end for l
      } # end for q 
    } # end if sparse
    
    mstep <- list(loglambda.ql=sol.hist$loglambda.ql,sum_rhotau_obs=sum_rhotau_obs,beta=sol.hist$beta)
    all.compl.log.likelihood[Qcand-Qmin+1] <- J.eval.mstep(VE,mstep,data,directed,sparse)$compl.log.lik
    
    # penalty term 
    sum.d.ql <- sum(sol.hist$best.d)
    all.pen[Qcand-Qmin+1] <- (Qcand-1)/2*log(n) + sum.d.ql/2*log(N)
    if (directed*sparse){ # penalizing beta values
      all.pen[Qcand-Qmin+1] <- all.pen[Qcand-Qmin+1]  + Qcand^2/2*log(N)  
    }    
    if ((!directed)*sparse){ # penalizing beta values
      all.pen[Qcand-Qmin+1] <- all.pen[Qcand-Qmin+1]  + Qcand*(Qcand+1)/4*log(N)}
    
    ICL <- all.compl.log.likelihood[Qcand-Qmin+1] - all.pen[Qcand-Qmin+1]
    all.ICL[Qcand-Qmin+1] <- ICL
    if (ICL>ICLbest){
      ICLbest <- ICL
      Qbest <- Qcand
      sol.Qbest <- sol.hist		
    }
  } # end loop for Qcand
  return(list(Qbest=Qbest, sol.Qbest=sol.Qbest, Qmin=Qmin,all.J=all.J,all.ICL=all.ICL,all.compl.log.likelihood=all.compl.log.likelihood,all.pen=all.pen))
}



###################################################
## Plots for Model selction
###################################################
# model.selec_Q : output from model.selection_Q()
plot.modelselec.Q <- function(model.selec_Q){
  Qmin <- model.selec_Q$Qmin
  Qmax <- length(model.selec_Q$all.ICL)+Qmin-1
  par(mfrow=c(1,2))
  ymin <- min(model.selec_Q$all.compl.log.likelihood,model.selec_Q$all.ICL)
  ymax <- max(model.selec_Q$all.compl.log.likelihood,model.selec_Q$all.ICL)
  plot(Qmin:Qmax,model.selec_Q$all.compl.log.likelihood,type='l',col='blue',xlab=paste('Qbest =',model.selec_Q$Qbest),ylab='ICL',ylim=c(ymin,ymax))
  lines(Qmin:Qmax,model.selec_Q$all.ICL,col='red')
  plot(Qmin:Qmax,model.selec_Q$all.J,ylab='J criterion',type='l',xlab='Q')
  return()
}


###################################################
## Computes direct kernel estimator densities relying on a classification tau
## This can be used with the values tau obtained on a dataset with dynppSBM fn run with 'hist' method
###################################################
#### warning sparse case not implemented !!! 
# data is a list with 3 components: 
#   data$time.seq : sequence of observed time points of the m-th event (M-vector)
#   data$type.seq : sequence of observed values which.dyads(i,j,n,directed) (auxiliary.R) of process that produced the mth event (M-vector). 
#   $Time - [0,data$Time] is the total time interval of observation 

hat.lambda.fn <- function(data,tau,Q,n,directed,sparse=FALSE,nb.points=1000){
  N <- if (directed) n*(n-1) else n*(n-1)/2
  N_Q <- if (directed) Q^2 else Q*(Q+1)/2
  
  if (sparse){#
    stop("Sorry sparse method not implemented yet")
    }
  
  step <- data$Time/nb.points
  partition <- seq(0,data$Time, by=step)
  hat.lambda.ql <- matrix(0,N_Q,length(partition))
  
  if (directed){ # directed case 
    ind.all <- which.index.all(n,directed)
    ind <- 0
    for (q in 1:Q){
      for (l in 1:Q){
        ind <- ind + 1
        if (Q==1){
          rhotauql <- rep(1,N)
          if (sparse){rhotauql[no.obs] <- rho}
        } else {
          rhotauql <- tau[q, ind.all[,1]]*tau[l, ind.all[,2]]  # N vector
          if (sparse){rhotauql[no.obs] <- rho[indice(q,l,Q,directed)]*rhotauql[no.obs]}
        }
        Yql <- sum(rhotauql)
        if (Yql>0){
          if (Q==1){
            poids <-rep(1/length(data$type.seq),length(data$type.seq))
            sum_tauql_obs <- length(data$type.seq)
          }	
          else {
            rhotau_obs <- rhotauql[data$type.seq]
            rhotau_obs_sum <- sum(rhotau_obs)
            #
            if (is.finite(sum(rhotau_obs/rhotau_obs_sum))){
              poids <- rhotau_obs/rhotau_obs_sum
            }
            else poids <- rep(0,length(rhotau_obs))
            dens.estim <- density(data$time.seq,weights=poids,kernel="epanechnikov")
            hat.lambda.ql[ind,] <- rhotau_obs_sum /Yql*(approxfun(dens.estim$x,dens.estim$y)(partition)) 
          }
        }
      } # end loop for l 
    } # end loop for q
  } else { # undirected case 
    ind.all <- which.index.all(n,directed)
    ind <- 0
    for (q in 1:Q){
      for (l in q:Q){
        ind <- ind + 1
        if (Q==1){
          rhotauql <- rep(1,N)
          rhotauql[no.obs] <- rho
        } else {
          if ((q==l)){
            rhotauql <- tau[q, ind.all[,1]]*tau[l, ind.all[,2]]  # N vector
            if (sparse){rhotauql[no.obs] <- rho[indice(q,l,Q,directed)]*rhotauql[no.obs]}
          } else {
            rhotauql <- tau[q, ind.all[,1]]*tau[l, ind.all[,2]] + tau[q, ind.all[,2]]*tau[l, ind.all[,1]]
            if (sparse){rhotauql[no.obs] <- rho[indice(q,l,Q,directed)]*rhotauql[no.obs]}
          }
        }
        Yql <- sum(rhotauql)
        if (Yql>0){
          if (Q==1){
            poids <-rep(1/length(data$type.seq),length(data$type.seq))
            sum_tauql_obs <- length(data$type.seq)
          }	
          else {
            rhotau_obs <- rhotauql[data$type.seq]
            rhotau_obs_sum <- sum(rhotau_obs)
            if (is.finite(sum(rhotau_obs/rhotau_obs_sum))){
              poids <- rhotau_obs/rhotau_obs_sum
            }
            else poids <- rep(0,length(rhotau_obs))
            dens.estim <- density(data$time.seq,weights=poids,kernel="epanechnikov")
            hat.lambda.ql[ind,] <- rhotau_obs_sum /Yql*(approxfun(dens.estim$x,dens.estim$y)(partition)) 
          }
        }
      } # end loop for l 
    } # end loop for q
  }
  return(hat.lambda.ql)
}



#######################################
### ADJUSTED RAND INDEX (ARI)
#######################################

randError<-function(x, y) {
  # function to calculate the adjusted rand statistic
  # x and y are vectors containing the two partitions to be compared
  # first, get crosstabs
  ctab <- table(x,y);
  # now calculate 4 intermediary sums
  cellsum <- sum(ctab*(ctab-1)/2)
  totsum <- sum(ctab)*(sum(ctab)-1)/2
  # use matrix multiplication to get row and column marginal sums
  rows <- ctab %*% rep(1,ncol(ctab))
  rowsum <- sum(rows*(rows-1)/2)
  cols <- rep(1,nrow(ctab)) %*% ctab
  colsum <- sum(cols*(cols-1)/2)
  # now put them together
  adj.rand <- (cellsum - (rowsum*colsum/totsum))/(.5*(rowsum +colsum)-(rowsum*colsum/totsum))
  return (adj.rand)
}


ARI <- function(z, hat.z){
  # z : 'true' labels, Q x n matrix with entries = 0 or 1
  # hat.z : estimated labels, Q x n matrix with 0<entries<1
  
  partition.z <- apply(z,2,which.max) 
  partition.hat.z <- apply(hat.z,2,which.max) 
  ARI <- randError(partition.z, partition.hat.z)
  return(ARI)
}

###################################################
##   COMPARISON OF THE Z ESTIMATORS WITH TRUE VALUES
###################################################
####
# computation of the permutation of the rows of hat.z to obtain the "same order" as z
####

# Compute optimal matching between 2 partitions using Hungarian algorithm
# z, hat.z : matrices of size Q x n
permute <- function(z,hat.z){ 
  map.z <- apply(z,2,which.max)
  map.hatz <- apply(hat.z,2,which.max)
  tab <- table(map.z,map.hatz)
  if (ncol(tab) < nrow(tab) ){ 
    res <- solve_LSAP(t(tab),maximum = TRUE)
  } else {res <- solve_LSAP(tab,maximum = TRUE)}
  return(res)
}


####
# sort lambda (associated with hat.z) "in the same way" as the original lambda (associated with z) by permutation of rows
####
permut.lambda <- function(lambda,z,hat.z,directed){
  Q <- nrow(z)
  perm <- permute(z,hat.z)
  
  # permutation of rows of lambda
  sol <- lambda
  for (q in 1:Q){
    vec.l <- if (directed) 1:Q else q:Q
    for (l in vec.l)
      if (!is.na(perm[q]*perm[l]))
        sol[indice(q,l,Q,directed),]<- lambda[indice(perm[q],perm[l],Q,directed),]
  }
  return(sol)
}



