######################################
## ANALYSIS OF RESULTS OF ENRON CORPUS
######################################

rm(list=ls())

# set path: data
# setwd("data")

# load the data
load("Enron_data.Rdata")

# load the solution of the VEM algorithm
load('Enron_solution.Rdata')  
sol_enron <- best

# source functions for the analysis of results
setwd("../code")
source("output.R")
setwd("../data")

# number of individuals
n <- 147

######################################
## Model selection via ICL 
######################################

Qmax <- length(sol_enron)
modelQhat <- model.selection_Q(data,n,1,Qmax,TRUE,FALSE,sol_enron)
plot.modelselec.Q(modelQhat)

######################################
## Analysis with Q = 4 groups
######################################

Q <- 4
solQ <- sol_enron[[Q]]

######################################
## logarithm of mean values of the estimated intensities alpha^(q,l)
######################################

library('ggplot2')
library('gridExtra')
library('reshape')
g_legend<-function(a.gplot){
  tmp <- ggplot_gtable(ggplot_build(a.gplot))
  leg <- which(sapply(tmp$grobs, function(x) x$name) == "guide-box")
  legend <- tmp$grobs[[leg]]
  return(legend)}


lambda.ql <- exp(solQ$loglambda.ql)
lambda.ql[lambda.ql==1] <- 0
lam_mean <- apply(lambda.ql,1,mean)*24*3600
log_lam_mean  <- log(lam_mean)
M <- matrix(log_lam_mean,Q,Q,byrow=T)
M <- round(M,2)
melted_M <- melt(M)
names(melted_M) <- c('q','l','log.alpha')
melted_M$q <- as.character(melted_M$q)
melted_M$l <- as.character(melted_M$l)
melted_M$q <- factor(melted_M$q,levels=as.character(1:Q))
melted_M$l <- factor(melted_M$l,levels=as.character(1:Q))

ggplot(data = melted_M, aes(x=l, y=q, fill=log.alpha)) +
  labs(x=expression(l),y=expression(q))+
  theme(text = element_text(size=25))+
  geom_tile(aes(fill = log.alpha)) +
  scale_fill_gradient2(low = "darkgreen", mid = "lightgreen", high='white') +
  geom_text(aes(l,q, label = log.alpha), color = "black", size = 10)

######################################
# Draw estimators of alpha^(q,l) and bootstrap confidence intervals  
######################################

load('Enron_bootstrap_results.RData')

# source functions for the analysis of results
setwd("../code")
source("bootstrap.R")
setwd("../data")

# compute confidence limits
CIlim80 <- boot.CIlimits(boot.sol,alpha=0.2)*24*60*60
CIlim90 <- boot.CIlimits(boot.sol,alpha=0.1)*24*60*60
alpha.hat <- exp(solQ$loglambda.ql)
alpha.hat[alpha.hat==1] <- 0
alpha.hat <- alpha.hat*24*60*60

tick1oct2001 <- 365*24*3600 + (31+30+31+31+30)*24*3600
tick1oct2000 <- 365*24*3600 + (31+30+31+31+30)*24*3600 - 365*24*3600
vec.tick <- c(tick1oct2000,tick1oct2001)
vec.lab <- c("01/10/00","01/10/01")
Dmax <- 2^8
vec.x <- (0:Dmax)*data$Time/Dmax

plot.list <-list()
for (q in 1:Q)
  for (l in 1:Q){
    ind.ql <- indice(q,l,Q, directed=T)
    data.plot<- melt(data.frame(x=vec.x,alpha=c(alpha.hat[ind.ql,],alpha.hat[ind.ql,Dmax])),id="x")
    levels(data.plot$variable) <- c("Intensity")
    L90 <- c(CIlim90[ind.ql,1,],CIlim90[ind.ql,1,Dmax])
    U90 <- c(CIlim90[ind.ql,2,],CIlim90[ind.ql,2,Dmax])
    L80 <- c(CIlim80[ind.ql,1,],CIlim80[ind.ql,1,Dmax])
    U80 <- c(CIlim80[ind.ql,2,],CIlim80[ind.ql,2,Dmax])
    data.polygon <- melt(data.frame(x=c(vec.x,rev(vec.x)), y80=c(L80,rev(U80)), y90=c(L90,rev(U90))),id="x")
    medianline <- c(CIlim80[ind.ql,3,],CIlim80[ind.ql,3,Dmax])
    data.median <- data.frame(x=vec.x,y=medianline)
    plot.list[[ind.ql]] <- ggplot(data =data.plot, aes(x=x,y=value,linetype = variable))+
      geom_line()+
      geom_polygon(data=data.polygon, aes(x=x,y=value, fill=variable),inherit.aes=F,alpha=0.5)+
      scale_fill_manual(values=c("darkgrey", "lightgrey")) +
      geom_line(data=data.median,aes(x=x,y=y,linetype='solid'))+
      guides(fill=FALSE)+ # removes legend for fill/levels
      scale_linetype_discrete(guide=FALSE) + #removes legend for linetype
      theme_bw()+ #beware: this should come before theme()
      theme(panel.grid.major = element_blank(), panel.grid.minor = element_blank(),
            panel.background = element_blank(), axis.line = element_line(colour = "black"),plot.title = element_text(hjust = 0.5))+ 
      labs(title =paste0("(q,l)=(",q,",",l,")"),y = "Intensity",x="Time") 
  }
# fake plot for legend
fake.plot <-  ggplot(data =data.plot, aes(x=x,y=value,linetype = variable))+
  geom_line() +
  geom_polygon(data=data.polygon, aes(x=x,y=value, fill=variable),inherit.aes=F,alpha=0.5)+
  scale_fill_manual(values=c("darkgrey", "lightgrey"),name="Level",labels=c("80%","90%")) +
  geom_line(data=data.median,aes(x=x,y=y,linetype='solid'))+
  scale_linetype_discrete(name="Intensities",labels=c("estimator","80% CI median"))+
  theme_bw()+
  labs(title =paste0("(q,l)=(",q,",",l,")"),y = "Intensity",x="Time") +
  theme(plot.title = element_text(hjust = 0.5),legend.position="bottom") 
  
mylegend <- g_legend(fake.plot)
interm.plot <- do.call("grid.arrange",c(plot.list, ncol=Q,nrow=Q))
final.plot <- grid.arrange(interm.plot, mylegend, nrow=2,heights=c(10,1))
ggsave("Enron_intens_all.pdf",plot=final.plot, width =8.5,height=5)

# Plot only estimates for pairs of groups (q,1)
plot.listq1 <- list(plot.list[[1]],plot.list[[5]],plot.list[[9]],plot.list[[13]])
interm.plot <- do.call("grid.arrange",c(plot.listq1, ncol=2,nrow=2))
final.plot <- grid.arrange(interm.plot, mylegend, nrow=2,heights=c(10,1))
ggsave("Enron_intens.pdf",plot=final.plot, width =5.8,height=4.5)


######################
## Analysis of positions at Enron
######################

# load data with the positions of the people at Enron
load('Enron_positions.RData') 

group.member <- apply(solQ$tau,2,which.max)
group.size <- table(group.member) 

# size of groups:
group.size
#1   2   3   4 
#127   4   2  14 

pos.counts <- matrix(0,3,Q)
list.manager <- vector("list", Q)
rownames(pos.counts) <- c("Employee","Manager","N/A")
for (i in 1:n){
  pos <- indiv[[i]][[3]]
  group.i <- group.member[i]
  if (pos == "N/A"){
    pos.counts[3,group.i] <- pos.counts[3,group.i] + 1   
  }else {
    if (pos == "Employee")
      pos.counts[1,group.i] <- pos.counts[1,group.i] + 1
    else{
      pos.counts[2,group.i] <- pos.counts[2,group.i] + 1
      list.manager[[group.i]] <- c(list.manager[[group.i]],pos)
    }
  }
}
for (k in 1:length(list.manager))
  if (length(list.manager[[k]])>0)
    list.manager[[k]] <- sort(list.manager[[k]])

# nb of employees/managers per group:
pos.counts
#[,1] [,2] [,3] [,4]
#Employee   36    3    1    1
#Manager    62    0    1   12
#N/A        29    1    0    1

# exact composition of manager positions per group:
list.manager

all.positions <- rep(NA,n)
for (i in 1:n)
  all.positions[i] <- indiv[[i]][[3]]

# total number of CEOs/Directors etc in the dataset:
table(all.positions)
# CEO          Director          Employee   In House Lawyer           Manager Managing Director               N/A         President            Trader    Vice President 
# 4                14                41                 1                14                 3                31                 4                13                22 


######################
###### Comparison to classical SBM 
######################

# load solution for classical SBM
load("Enron_solution_SBM.Rdata")
sol_SBM <- best
# load data for classical SBM
load("Enron_data_SBM.Rdata")

## Model selection via ICL
Qmax <- 20
modelQ_sbm <-model.selection_Q(data,n,1,Qmax,TRUE,FALSE,sol_SBM)
plot.modelselec.Q(modelQ_sbm)
# Q = 17 is the optimal number of groups according to ICL

# analysis with Q=4 groups and comparison to PPSBM solution with Q=4
Q <- 4
group_sbm <- apply(sol_SBM[[Q]]$tau,2,which.max)

# compare group sizes of the cluterings in the two models (SBM and PPSBM):
table(group_sbm)
# 1   2   3   4 
# 123  16   3   5 
table(group.member)
# 1   2   3   4 
# 127   4   2  14 

# compute ARI of the two clusterings
ARI(solQ$tau,sol_SBM[[Q]]$tau)
# 0.7982786


# compare the composition of the different groups in the two clusterings:
length(intersect((1:n)[group.member==1],(1:n)[group_sbm==1]))
# 122
setdiff((1:n)[group.member==1],(1:n)[group_sbm==1])
#  2 35 47 59 71
setdiff((1:n)[group_sbm==1],(1:n)[group.member==1])
# 8

(1:n)[group.member==2]
#[1]  8 36 73 74
(1:n)[group_sbm==4]
#[1]  1 36 53 59 67
intersect((1:n)[group.member==2],(1:n)[group_sbm==4])
# 36

(1:n)[group.member==4]
#[1]  1  6 18 21 42 43 44 48 49 53 60 68 69 72
(1:n)[group_sbm==2]
#[1]  2  6 18 21 35 42 44 47 48 49 60 68 69 71 72 74
length(intersect((1:n)[group.member==4],(1:n)[group_sbm==2]))
# 11

(1:n)[group.member==3]
#[1] 19 67
(1:n)[group_sbm==3]
#[1] 19 43 73
intersect((1:n)[group.member==3],(1:n)[group_sbm==3])
# 19

#############################
## Sparse model
#############################

## load solution
load('Enron_solution_sparse.Rdata')
sol_sparse <- best

# reload the data
load("Enron_data.Rdata")

## Model selection via ICL
modelQsparse <- model.selection_Q(data,n,Qmin=1,Qmax=20,directed=TRUE,sparse=TRUE,sol_sparse)
plot.modelselec.Q(modelQsparse)
# best Q : 10


# Comparison to non sparse model at Q=10
Q <- 4
solQsparse <- sol_sparse[[Q]]

# compute ARI to compare the clustering in the non sparse model
ARI(solQ$tau,solQsparse$tau)
# [1] 0.5193341

# Comparison to non sparse model at Q=10
Q <- 10
solQsparse <- sol_sparse[[Q]]

# group sizes
table(apply(solQsparse$tau,2,which.max))
# 1   2   3   4   5   6   7   8   9  10 
# 1 125   1   1   2   3   5   3   2   4 

# estimated connectivity parameters beta^(q,l)
beta <- matrix(round(solQsparse$beta,1),Q,Q,byrow=T)
melted_beta <- melt(beta)
names(melted_beta) <- c('q','l','beta')
melted_beta$q <- as.character(melted_beta$q)
melted_beta$l <- as.character(melted_beta$l)
melted_beta$q <- factor(melted_beta$q,levels=as.character(1:Q))
melted_beta$l <- factor(melted_beta$l,levels=as.character(1:Q))

ggplot(data = melted_beta, aes(x=l, y=q, fill=beta)) +
  theme(text = element_text(size=25))+
  geom_tile(aes(fill = beta)) +
  scale_fill_gradient2(low = "black", mid = "lightblue", high='white') +
  geom_text(aes(l,q, label = beta), color = "black", size = 10)


# log of mean intensities alpha^(q,l)
lambda.ql <- exp(solQsparse$loglambda.ql)
lambda.ql[lambda.ql==1] <- 0
lam_mean_sparse <- apply(lambda.ql,1,mean)*24*3600
log_lam_mean  <- log(lam_mean_sparse)
Msparse <- matrix(log_lam_mean,Q,Q,byrow=T)
Msparse <- round(Msparse,1)
melted_Msparse <- melt(Msparse)
names(melted_Msparse) <- c('q','l','log.alpha')
melted_Msparse$q <- as.character(melted_Msparse$q)
melted_Msparse$l <- as.character(melted_Msparse$l)
melted_Msparse$q <- factor(melted_Msparse$q,levels=as.character(1:Q))
melted_Msparse$l <- factor(melted_Msparse$l,levels=as.character(1:Q))

ggplot(data = melted_Msparse, aes(x=l, y=q, fill=log.alpha)) +
  theme(text = element_text(size=25))+
  geom_tile(aes(fill = log.alpha)) +
  scale_fill_gradient2(low = "darkgreen", mid = "lightgreen", high='white') +
  geom_text(aes(l,q, label = log.alpha), color = "black", size = 10)

# compute ARI to compare the clustering in the non sparse model
ARI(solQ$tau,solQsparse$tau)
# [1] 0.5126217


# percentage of pairs of individuals that do never communicate
mean(rowSums(data$Nijk)==0)
# [1] 0.9118908

nb.proc.ql <- matrix(NA,Q,Q)
ind.all <- which.index.all(n)
ind.ql <- 0
for (q in 1:Q){
  for (l in 1:Q){
    ind.ql <- ind.ql + 1
    tauql <- solQsparse$tau[q, ind.all[,1]]*solQsparse$tau[l, ind.all[,2]] # N-vector
    nb.proc.ql[q,l] <- sum(tauql)
  }
}
nb.proc.ql    


sum(nb.proc.ql*solQsparse$beta)
# [1] 1928.23
sum(rowSums(data$Nijk)!=0)
# [1] 1891



alpha.sparse <- exp(solQsparse$loglambda.ql)
alpha.sparse[alpha.sparse==1] <- 0
alpha.sparse <- alpha.sparse*24*60*60
#par(mfrow=c(2,5))
plot.list<-list()
l <- 2
i<-0
for (q in 1:10){
  i<-i+1
  ind.ql <- indice(q,l,Q, directed=T)
  data.plot <- data.frame(Time=vec.x,Intensity=c(alpha.sparse[ind.ql,],alpha.sparse[ind.ql,Dmax]))
  plot.list[[i]] <- ggplot(data.plot,aes(x=Time,y=Intensity))+
    geom_line() +
    labs(title =paste0("(q,l)=(",q,",",l,")")) +
    theme(plot.title = element_text(hjust = 0.5)) +
    scale_x_continuous(breaks=vec.tick,labels = vec.lab)
}
final.plot <- do.call("grid.arrange",c(plot.list, ncol=3,nrow=4))


# Comparison to non sparse model at Q=4
Q <- 4
solQsparse <- sol_sparse[[Q]]

# group sizes
table(apply(solQsparse$tau,2,which.max))
# 1   2   3   4 
# 126   4   3  14 

# estimated connectivity parameters beta^(q,l)
beta <- matrix(round(solQsparse$beta,2),Q,Q,byrow=T)
melted_beta <- melt(beta)
names(melted_beta) <- c('q','l','beta')
melted_beta$q <- as.character(melted_beta$q)
melted_beta$l <- as.character(melted_beta$l)
melted_beta$q <- factor(melted_beta$q,levels=as.character(1:Q))
melted_beta$l <- factor(melted_beta$l,levels=as.character(1:Q))

ggplot(data = melted_beta, aes(x=l, y=q, fill=beta)) +
  theme(text = element_text(size=25))+
  geom_tile(aes(fill = beta)) +
  scale_fill_gradient2(low = "black", mid = "lightblue", high='white') +
  geom_text(aes(l,q, label = beta), color = "black", size = 10)


# log of mean intensities alpha^(q,l)
lambda.ql <- exp(solQsparse$loglambda.ql)
lambda.ql[lambda.ql==1] <- 0
lam_mean_sparse <- apply(lambda.ql,1,mean)*24*3600
log_lam_mean  <- log(lam_mean_sparse)
Msparse <- matrix(log_lam_mean,Q,Q,byrow=T)
Msparse <- round(Msparse,2)
melted_Msparse <- melt(Msparse)
names(melted_Msparse) <- c('q','l','log.alpha')
melted_Msparse$q <- as.character(melted_Msparse$q)
melted_Msparse$l <- as.character(melted_Msparse$l)
melted_Msparse$q <- factor(melted_Msparse$q,levels=as.character(1:Q))
melted_Msparse$l <- factor(melted_Msparse$l,levels=as.character(1:Q))

ggplot(data = melted_Msparse, aes(x=l, y=q, fill=log.alpha)) +
  theme(text = element_text(size=25))+
  geom_tile(aes(fill = log.alpha)) +
  scale_fill_gradient2(low = "darkgreen", mid = "lightgreen", high='white') +
  geom_text(aes(l,q, label = log.alpha), color = "black", size = 10)

## intensities are much more elevated than in the non sparse case:
# the difference is approximately beta:
matrix(round(lam_mean/lam_mean_sparse,3),Q,Q,byrow = T)
beta <- matrix(solQsparse$beta,Q,Q,byrow=T)
round(beta,3)

# percentage of pairs of individuals that do never communicate
mean(rowSums(data$Nijk)==0)
# [1] 0.9118908

nb.proc.ql <- matrix(NA,Q,Q)
ind.all <- which.index.all(n)
ind.ql <- 0
for (q in 1:Q){
  for (l in 1:Q){
    ind.ql <- ind.ql + 1
    tauql <- solQsparse$tau[q, ind.all[,1]]*solQsparse$tau[l, ind.all[,2]] # N-vector
    nb.proc.ql[q,l] <- sum(tauql)
  }
}
nb.proc.ql    

sum(nb.proc.ql*solQsparse$beta)
# [1] 1892.298
sum(rowSums(data$Nijk)!=0)
# [1] 1891


# Draw estimators of alpha^(q,l)   
alpha.sparse <- exp(solQsparse$loglambda.ql)
alpha.sparse[alpha.sparse==1] <- 0
alpha.sparse <- alpha.sparse*24*60*60
plot.list<-list()
for (q in 1:Q)
  for (l in 1:Q){
    ind.ql <- indice(q,l,Q, directed=T)
    data.plot <- data.frame(Time=vec.x,Intensity=c(alpha.sparse[ind.ql,],alpha.sparse[ind.ql,Dmax]))
    plot.list[[ind.ql]] <- ggplot(data.plot,aes(x=Time,y=Intensity))+
      geom_line() +
      labs(title =paste0("(q,l)=(",q,",",l,")")) +
      theme(plot.title = element_text(hjust = 0.5)) +
      scale_x_continuous(breaks=vec.tick,labels = vec.lab)
  }
final.plot <- do.call("grid.arrange",c(plot.list, ncol=4,nrow=4))
ggsave("Enron_sparse_intens_Q4.pdf",final.plot, width =8.5,height=5)
